#!/usr/bin/env python3
#
#
# Run librdkafka regression tests on different supported broker versions.
#
# Requires:
#  trivup python module
#  gradle in your PATH

from cluster_testing import read_scenario_conf
from broker_version_tests import test_it

import os
import sys
import argparse
import json


def version_as_number(version):
    if version == 'trunk':
        return sys.maxsize
    tokens = version.split('.')
    return float('%s.%s' % (tokens[0], tokens[1]))


def test_version(version, cmd=None, deploy=True, conf={}, debug=False,
                 exec_cnt=1,
                 root_path='tmp', broker_cnt=3, scenario='default',
                 kraft=False):
    """
    @brief Create, deploy and start a Kafka cluster using Kafka \\p version
    Then run librdkafka's regression tests. Use inherited environment.
    """
    conf['test_mode'] = 'bash'
    test_it(version, deploy, conf, {}, None, True, debug,
            scenario, kraft, True)
    return True


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description='Start a Kafka cluster and provide an interactive shell')

    parser.add_argument('versions', type=str, default=None, nargs='+',
                        help='Kafka version(s) to deploy')
    parser.add_argument('--no-deploy', action='store_false', dest='deploy',
                        default=True,
                        help='Dont deploy applications, '
                        'assume already deployed.')
    parser.add_argument('--conf', type=str, dest='conf', default=None,
                        help='''
    JSON config object (not file).
    This does not translate to broker configs directly.
    If broker config properties are to be specified,
    they should be specified with
    --conf \'{"conf": ["key=value", "key=value"]}\'''')
    parser.add_argument('--scenario', type=str, dest='scenario',
                        default='default',
                        help='Test scenario (see scenarios/ directory)')
    parser.add_argument('-c', type=str, dest='cmd', default=None,
                        help='Command to execute instead of shell')
    parser.add_argument('-n', type=int, dest='exec_cnt', default=1,
                        help='Number of times to execute -c ..')
    parser.add_argument('--debug', action='store_true', dest='debug',
                        default=False,
                        help='Enable trivup debugging')
    parser.add_argument(
        '--root',
        type=str,
        default=os.environ.get(
            'TRIVUP_ROOT',
            'tmp'),
        help='Root working directory')
    parser.add_argument(
        '--port',
        default=None,
        help='Base TCP port to start allocating from')
    parser.add_argument(
        '--kafka-src',
        dest='kafka_path',
        type=str,
        default=None,
        help='Path to Kafka git repo checkout (used for version=trunk)')
    parser.add_argument(
        '--brokers',
        dest='broker_cnt',
        type=int,
        default=3,
        help='Number of Kafka brokers')
    parser.add_argument('--ssl', dest='ssl', action='store_true',
                        default=False,
                        help='Enable SSL endpoints')
    parser.add_argument(
        '--sasl',
        dest='sasl',
        type=str,
        default=None,
        help='SASL mechanism (PLAIN, SCRAM-SHA-nnn, GSSAPI, OAUTHBEARER)')
    parser.add_argument(
        '--oauthbearer-method',
        dest='sasl_oauthbearer_method',
        type=str,
        default=None,
        help='OAUTHBEARER/OIDC method (DEFAULT, OIDC), \
             must config SASL mechanism to OAUTHBEARER')
    parser.add_argument(
        '--max-reauth-ms',
        dest='reauth_ms',
        type=int,
        default='10000',
        help='''
        Sets the value of connections.max.reauth.ms on the brokers.
        Set 0 to disable.''')
    parser.add_argument(
        '--kraft',
        dest='kraft',
        action='store_true',
        default=False,
        help='Run in KRaft mode')

    args = parser.parse_args()
    if args.conf is not None:
        args.conf = json.loads(args.conf)
    else:
        args.conf = {}

    args.conf.update(read_scenario_conf(args.scenario))

    if args.port is not None:
        args.conf['port_base'] = int(args.port)
    if args.kafka_path is not None:
        args.conf['kafka_path'] = args.kafka_path
    if args.ssl:
        args.conf['security.protocol'] = 'SSL'
    if args.sasl:
        if (args.sasl == 'PLAIN' or args.sasl.find('SCRAM')
                != -1) and 'sasl_users' not in args.conf:
            args.conf['sasl_users'] = 'testuser=testpass'
        args.conf['sasl_mechanisms'] = args.sasl
    retcode = 0
    if args.sasl_oauthbearer_method:
        if args.sasl_oauthbearer_method == "OIDC" and \
           args.conf['sasl_mechanisms'] != 'OAUTHBEARER':
            print('If config `--oauthbearer-method=OIDC`, '
                  '`--sasl` must be set to `OAUTHBEARER`')
            retcode = 3
            sys.exit(retcode)
        args.conf['sasl_oauthbearer_method'] = \
            args.sasl_oauthbearer_method

    if 'conf' not in args.conf:
        args.conf['conf'] = []

    args.conf['conf'].append(
        "connections.max.reauth.ms={}".format(
            args.reauth_ms))
    args.conf['conf'].append("log.retention.bytes=1000000000")

    for version in args.versions:
        r = test_version(version, cmd=args.cmd, deploy=args.deploy,
                         conf=args.conf, debug=args.debug,
                         exec_cnt=args.exec_cnt,
                         root_path=args.root, broker_cnt=args.broker_cnt,
                         scenario=args.scenario,
                         kraft=args.kraft)
        if not r:
            retcode = 2

    sys.exit(retcode)



#!/usr/bin/env python3
#
#
# Run librdkafka regression tests on with different SASL parameters
# and broker verisons.
#
# Requires:
#  trivup python module
#  gradle in your PATH

from cluster_testing import (
    LibrdkafkaTestCluster,
    print_report_summary,
    read_scenario_conf)
from LibrdkafkaTestApp import LibrdkafkaTestApp

import subprocess
import tempfile
import os
import sys
import argparse
import json


def test_it(version, deploy=True, conf={}, rdkconf={}, tests=None,
            interact=False, debug=False, scenario="default", kraft=False,
            inherit_env=False):
    """
    @brief Create, deploy and start a Kafka cluster using Kafka \\p version
    Then run librdkafka's regression tests.
    """

    cluster = LibrdkafkaTestCluster(version, conf,
                                    num_brokers=int(conf.get('broker_cnt', 3)),
                                    debug=debug, scenario=scenario,
                                    kraft=kraft)

    # librdkafka's regression tests, as an App.
    _rdkconf = conf.copy()  # Base rdkconf on cluster conf + rdkconf
    _rdkconf.update(rdkconf)
    rdkafka = LibrdkafkaTestApp(cluster, version, _rdkconf, tests=tests,
                                scenario=scenario)
    rdkafka.do_cleanup = False

    if deploy:
        cluster.deploy()

    cluster.start(timeout=30)

    if conf.get('test_mode', '') == 'bash':
        rdkafka.finalize_env()

        if inherit_env:
            env = dict(os.environ, **rdkafka.env)
        else:
            env = dict(rdkafka.env)
        trivup = f'[TRIVUP:{cluster.name}@{version}] '
        PS1 = ((trivup + env['PS1']) if 'PS1' in env
               else trivup + '\\u@\\h:\\w$ ')\
            .translate(str.maketrans({'\'': '\\\''}))
        cmd = f'bash --rcfile <(cat ~/.bashrc; echo \'PS1="{PS1}"\')'
        subprocess.call(
            cmd,
            env=env,
            shell=True,
            executable='/bin/bash')
        report = None

    else:
        rdkafka.start()
        print(
            '# librdkafka regression tests started, logs in %s' %
            rdkafka.root_path())
        rdkafka.wait_stopped(timeout=60 * 30)

        report = rdkafka.report()
        report['root_path'] = rdkafka.root_path()

        if report.get('tests_failed', 0) > 0 and interact:
            print(
                '# Connect to cluster with bootstrap.servers %s' %
                cluster.bootstrap_servers())
            print('# Exiting the shell will bring down the cluster. '
                  'Good luck.')
            subprocess.call(
                'bash --rcfile <(cat ~/.bashrc; echo \'PS1="[TRIVUP:%s@%s] \\u@\\h:\\w$ "\')' %  # noqa: E501
                (cluster.name, version), env=rdkafka.env, shell=True,
                executable='/bin/bash')

    cluster.stop(force=True)

    cluster.cleanup()
    return report


def handle_report(report, version, suite):
    """ Parse test report and return tuple (Passed(bool), Reason(str)) """
    test_cnt = report.get('tests_run', 0)

    if test_cnt == 0:
        return (False, 'No tests run')

    passed = report.get('tests_passed', 0)
    failed = report.get('tests_failed', 0)
    if 'all' in suite.get('expect_fail', []) or version in suite.get(
            'expect_fail', []):
        expect_fail = True
    else:
        expect_fail = False

    if expect_fail:
        if failed == test_cnt:
            return (True, 'All %d/%d tests failed as expected' %
                    (failed, test_cnt))
        else:
            return (False, '%d/%d tests failed: expected all to fail' %
                    (failed, test_cnt))
    else:
        if failed > 0:
            return (False, '%d/%d tests passed: expected all to pass' %
                    (passed, test_cnt))
        else:
            return (True, 'All %d/%d tests passed as expected' %
                    (passed, test_cnt))


if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description='Run librdkafka tests on a range of broker versions')

    parser.add_argument('--debug', action='store_true', default=False,
                        help='Enable trivup debugging')
    parser.add_argument('--conf', type=str, dest='conf', default=None,
                        help='trivup JSON config object (not file)')
    parser.add_argument('--rdkconf', type=str, dest='rdkconf', default=None,
                        help='trivup JSON config object (not file) '
                        'for LibrdkafkaTestApp')
    parser.add_argument('--scenario', type=str, dest='scenario',
                        default='default',
                        help='Test scenario (see scenarios/ directory)')
    parser.add_argument('--tests', type=str, dest='tests', default=None,
                        help='Test to run (e.g., "0002")')
    parser.add_argument('--report', type=str, dest='report', default=None,
                        help='Write test suites report to this filename')
    parser.add_argument('--interact', action='store_true', dest='interact',
                        default=False,
                        help='On test failure start a shell before bringing '
                        'the cluster down.')
    parser.add_argument('versions', type=str, nargs='*',
                        default=['0.8.1.1', '0.8.2.2', '0.9.0.1', '2.3.0'],
                        help='Broker versions to test')
    parser.add_argument('--interactive', action='store_true',
                        dest='interactive',
                        default=False,
                        help='Start a shell instead of running tests')
    parser.add_argument(
        '--root',
        type=str,
        default=os.environ.get(
            'TRIVUP_ROOT',
            'tmp'),
        help='Root working directory')
    parser.add_argument(
        '--port',
        default=None,
        help='Base TCP port to start allocating from')
    parser.add_argument(
        '--kafka-src',
        dest='kafka_path',
        type=str,
        default=None,
        help='Path to Kafka git repo checkout (used for version=trunk)')
    parser.add_argument(
        '--brokers',
        dest='broker_cnt',
        type=int,
        default=3,
        help='Number of Kafka brokers')
    parser.add_argument('--ssl', dest='ssl', action='store_true',
                        default=False,
                        help='Enable SSL endpoints')
    parser.add_argument(
        '--sasl',
        dest='sasl',
        type=str,
        default=None,
        help='SASL mechanism (PLAIN, GSSAPI)')
    parser.add_argument(
        '--kraft',
        dest='kraft',
        action='store_true',
        default=False,
        help='Run in KRaft mode')

    args = parser.parse_args()

    conf = dict()
    rdkconf = dict()

    if args.conf is not None:
        args.conf = json.loads(args.conf)
    else:
        args.conf = {}

    if args.port is not None:
        args.conf['port_base'] = int(args.port)
    if args.kafka_path is not None:
        args.conf['kafka_path'] = args.kafka_path
    if args.ssl:
        args.conf['security.protocol'] = 'SSL'
    if args.sasl:
        if args.sasl == 'PLAIN' and 'sasl_users' not in args.conf:
            args.conf['sasl_users'] = 'testuser=testpass'
        args.conf['sasl_mechanisms'] = args.sasl
        args.conf['sasl_servicename'] = 'kafka'
    if args.interactive:
        args.conf['test_mode'] = 'bash'
    args.conf['broker_cnt'] = args.broker_cnt

    conf.update(args.conf)
    if args.rdkconf is not None:
        rdkconf.update(json.loads(args.rdkconf))

    conf.update(read_scenario_conf(args.scenario))

    if args.tests is not None:
        tests = args.tests.split(',')
    elif 'tests' in conf:
        tests = conf.get('tests', '').split(',')
    else:
        tests = None

    # Test version + suite matrix
    if 'versions' in conf:
        versions = conf.get('versions')
    else:
        versions = args.versions
    suites = [{'name': 'standard'}]

    pass_cnt = 0
    fail_cnt = 0
    for version in versions:
        for suite in suites:
            _conf = conf.copy()
            _conf.update(suite.get('conf', {}))
            _rdkconf = rdkconf.copy()
            _rdkconf.update(suite.get('rdkconf', {}))

            if 'version' not in suite:
                suite['version'] = dict()

            # Run tests
            print('#### Version %s, suite %s, scenario %s: STARTING' %
                  (version, suite['name'], args.scenario))
            report = test_it(version, tests=tests, conf=_conf,
                             rdkconf=_rdkconf,
                             interact=args.interact, debug=args.debug,
                             scenario=args.scenario,
                             kraft=args.kraft)

            if not report:
                continue

            # Handle test report
            report['version'] = version
            passed, reason = handle_report(report, version, suite)
            report['PASSED'] = passed
            report['REASON'] = reason

            if passed:
                print('\033[42m#### Version %s, suite %s: PASSED: %s\033[0m' %
                      (version, suite['name'], reason))
                pass_cnt += 1
            else:
                print('\033[41m#### Version %s, suite %s: FAILED: %s\033[0m' %
                      (version, suite['name'], reason))
                fail_cnt += 1

                # Emit hopefully relevant parts of the log on failure
                subprocess.call(
                    "grep --color=always -B100 -A10 FAIL %s" %
                    (os.path.join(
                        report['root_path'],
                        'stderr.log')),
                    shell=True)

            print('#### Test output: %s/stderr.log' % (report['root_path']))

            suite['version'][version] = report

    # Write test suite report JSON file
    if args.report is not None:
        test_suite_report_file = args.report
        f = open(test_suite_report_file, 'w')
    else:
        fd, test_suite_report_file = tempfile.mkstemp(prefix='test_suite_',
                                                      suffix='.json',
                                                      dir='.')
        f = os.fdopen(fd, 'w')

    full_report = {'suites': suites, 'pass_cnt': pass_cnt,
                   'fail_cnt': fail_cnt, 'total_cnt': pass_cnt + fail_cnt}

    f.write(json.dumps(full_report))
    f.close()

    print('\n\n\n')
    print_report_summary(full_report)
    print('#### Full test suites report in: %s' % test_suite_report_file)

    if pass_cnt == 0 or fail_cnt > 0:
        sys.exit(1)
    else:
        sys.exit(0)



#!/usr/bin/env python3
#
# Packaging script.
# Assembles packages using CI artifacts.
#

import sys
import re
import os
import shutil
from fnmatch import fnmatch
from string import Template
from zfile import zfile
import boto3
import magic

if sys.version_info[0] < 3:
    from urllib import unquote as _unquote
else:
    from urllib.parse import unquote as _unquote


def unquote(path):
    # Removes URL escapes, and normalizes the path by removing ./.
    path = _unquote(path)
    if path[:2] == './':
        return path[2:]
    return path


# Rename token values
rename_vals = {'plat': {'windows': 'win'},
               'arch': {'x86_64': 'x64',
                        'amd64': 'x64',
                        'i386': 'x86',
                        'win32': 'x86'}}

# Filemagic arch mapping.
# key is (plat, arch, file_extension), value is a compiled filemagic regex.
# This is used to verify that an artifact has the expected file type.
magic_patterns = {
    ('win', 'x64', '.dll'): re.compile('PE32.*DLL.* x86-64, for MS Windows'),
    ('win', 'x86', '.dll'):
    re.compile('PE32.*DLL.* Intel 80386, for MS Windows'),
    ('win', 'x64', '.lib'): re.compile('current ar archive'),
    ('win', 'x86', '.lib'): re.compile('current ar archive'),
    ('linux', 'x64', '.so'): re.compile('ELF 64.* x86-64'),
    ('linux', 'arm64', '.so'): re.compile('ELF 64.* ARM aarch64'),
    ('osx', 'x64', '.dylib'): re.compile('Mach-O 64.* x86_64'),
    ('osx', 'arm64', '.dylib'): re.compile('Mach-O 64.*arm64')}

magic = magic.Magic()


def magic_mismatch(path, a):
    """ Verify that the filemagic for \\p path matches for artifact \\p a.
        Returns True if the magic file info does NOT match.
        Returns False if no matching is needed or the magic matches. """
    k = (a.info.get('plat', None), a.info.get('arch', None),
         os.path.splitext(path)[1])
    pattern = magic_patterns.get(k, None)
    if pattern is None:
        return False

    minfo = magic.id_filename(path)
    if not pattern.match(minfo):
        print(
            f"Warning: {path} magic \"{minfo}\" "
            f"does not match expected {pattern} for key {k}")
        return True

    return False


# Collects CI artifacts from S3 storage, downloading them
# to a local directory, or collecting already downloaded artifacts from
# local directory.
#
# The artifacts' folder in the S3 bucket must have the following token
# format:
#  <token>-[<value>]__   (repeat)
#
# Recognized tokens (unrecognized tokens are ignored):
#  p       - project (e.g., "confluent-kafka-python")
#  bld     - builder (e.g., "travis")
#  plat    - platform ("osx", "linux", ..)
#  dist    - distro or runtime ("centos8", "mingw", "msvcr", "alpine", ..).
#  arch    - arch ("x64", ..)
#  tag     - git tag
#  sha     - git sha
#  bid     - builder's build-id
#  bldtype - Release, Debug (appveyor)
#  lnk     - Linkage ("std", "static", "all" (both std and static))
#  extra   - Extra build options, typically "gssapi" (for cyrus-sasl linking).

#
# Example:
#   librdkafka/p-librdkafka__bld-travis__plat-linux__arch-x64__tag-v0.0.62__sha-d051b2c19eb0c118991cd8bc5cf86d8e5e446cde__bid-1562.1/librdkafka.tar.gz


class MissingArtifactError(Exception):
    pass


s3_bucket = 'librdkafka-ci-packages'
dry_run = False


class Artifact (object):
    def __init__(self, arts, path, info=None):
        self.path = path
        # Remove unexpanded AppVeyor $(..) tokens from filename
        self.fname = re.sub(r'\$\([^\)]+\)', '', os.path.basename(path))
        slpath = os.path.join(os.path.dirname(path), self.fname)
        if os.path.isfile(slpath):
            # Already points to local file in correct location
            self.lpath = slpath
        else:
            # Prepare download location in dlpath
            self.lpath = os.path.join(arts.dlpath, slpath)

        if info is None:
            self.info = dict()
        else:
            # Assign the map and convert all keys to lower case
            self.info = {k.lower(): v for k, v in info.items()}
            # Rename values, e.g., 'plat':'windows' to 'plat':'win'
            for k, v in self.info.items():
                rdict = rename_vals.get(k, None)
                if rdict is not None:
                    self.info[k] = rdict.get(v, v)

        # Score value for sorting
        self.score = 0

        # AppVeyor symbol builds are of less value
        if self.fname.find('.symbols.') != -1:
            self.score -= 10

        self.arts = arts
        arts.artifacts.append(self)

    def __repr__(self):
        return self.path

    def __lt__(self, other):
        return self.score < other.score

    def download(self):
        """ Download artifact from S3 and store in local directory .lpath.
            If the artifact is already downloaded nothing is done. """
        if os.path.isfile(self.lpath) and os.path.getsize(self.lpath) > 0:
            return
        print('Downloading %s' % self.path)
        if dry_run:
            return
        ldir = os.path.dirname(self.lpath)
        if not os.path.isdir(ldir):
            os.makedirs(ldir, 0o755)
        self.arts.s3_bucket.download_file(self.path, self.lpath)


class Artifacts (object):
    def __init__(self, match, dlpath):
        super(Artifacts, self).__init__()
        self.match = match
        self.artifacts = list()
        # Download directory (make sure it ends with a path separator)
        if not dlpath.endswith(os.path.sep):
            dlpath = os.path.join(dlpath, '')
        self.dlpath = dlpath
        if not os.path.isdir(self.dlpath):
            if not dry_run:
                os.makedirs(self.dlpath, 0o755)

    def collect_single(self, path, req_tag=True):
        """ Collect single artifact, be it in S3 or locally.
        :param: path string: S3 or local (relative) path
        :param: req_tag bool: Require tag to match.
        """

        # For local files, strip download path.
        # Also ignore any parent directories.
        if path.startswith(self.dlpath):
            folder = os.path.basename(os.path.dirname(path[len(self.dlpath):]))
        else:
            folder = os.path.basename(os.path.dirname(path))

        # The folder contains the tokens needed to perform
        # matching of project, gitref, etc.
        rinfo = re.findall(r'(?P<tag>[^-]+)-(?P<val>.*?)(?:__|$)', folder)
        if rinfo is None or len(rinfo) == 0:
            print('Incorrect folder/file name format for %s' % folder)
            return None

        info = dict(rinfo)

        # Ignore AppVeyor Debug builds
        if info.get('bldtype', '').lower() == 'debug':
            print('Ignoring debug artifact %s' % folder)
            return None

        tag = info.get('tag', None)
        if tag is not None and (len(tag) == 0 or tag.startswith('$(')):
            # AppVeyor doesn't substite $(APPVEYOR_REPO_TAG_NAME)
            # with an empty value when not set, it leaves that token
            # in the string - so translate that to no tag.
            del info['tag']

        # Perform matching
        unmatched = list()
        for m, v in self.match.items():
            if m not in info or info[m] != v:
                unmatched.append(f"{m} = {v}")

        # Make sure all matches were satisfied, unless this is a
        # common artifact.
        if info.get('p', '') != 'common' and len(unmatched) > 0:
            return None

        return Artifact(self, path, info)

    def collect_s3(self):
        """ Collect and download build-artifacts from S3 based on
        git reference """
        print(
            'Collecting artifacts matching %s from S3 bucket %s' %
            (self.match, s3_bucket))
        self.s3 = boto3.resource('s3')
        self.s3_bucket = self.s3.Bucket(s3_bucket)
        self.s3_client = boto3.client('s3')

        # note: list_objects will return at most 1000 objects per call,
        #       use continuation token to read full list.
        cont_token = None
        more = True
        while more:
            if cont_token is not None:
                res = self.s3_client.list_objects_v2(
                    Bucket=s3_bucket,
                    Prefix='librdkafka/',
                    ContinuationToken=cont_token)
            else:
                res = self.s3_client.list_objects_v2(Bucket=s3_bucket,
                                                     Prefix='librdkafka/')

            if res.get('IsTruncated') is True:
                cont_token = res.get('NextContinuationToken')
            else:
                more = False

            for item in res.get('Contents'):
                self.collect_single(item.get('Key'))

        for a in self.artifacts:
            a.download()

    def collect_local(self, path, req_tag=True):
        """ Collect artifacts from a local directory possibly previously
        collected from s3 """
        for f in [os.path.join(dp, f) for dp, dn,
                  filenames in os.walk(path) for f in filenames]:
            if not os.path.isfile(f):
                continue
            self.collect_single(f, req_tag)


class Mapping (object):
    """ Maps/matches a file in an input release artifact to
        the output location of the package, based on attributes and paths. """

    def __init__(self, attributes, artifact_fname_glob, path_in_artifact,
                 output_pkg_path=None, artifact_fname_excludes=[]):
        """
        @param attributes A dict of artifact attributes that must match.
                          If an attribute name (dict key) is prefixed
                          with "!" (e.g., "!plat") then the attribute
                          must not match.
        @param artifact_fname_glob Match artifacts with this filename glob.
        @param path_in_artifact On match, extract this file in the artifact,..
        @param output_pkg_path ..and write it to this location in the package.
                               Defaults to path_in_artifact.
        @param artifact_fname_excludes Exclude artifacts matching these
                                       filenames.

        Pass a list of Mapping objects to FIXME to perform all mappings.
        """
        super(Mapping, self).__init__()
        self.attributes = attributes
        self.fname_glob = artifact_fname_glob
        self.input_path = path_in_artifact
        if output_pkg_path is None:
            self.output_path = self.input_path
        else:
            self.output_path = output_pkg_path
        self.name = self.output_path
        self.fname_excludes = artifact_fname_excludes

    def __str__(self):
        return self.name


class Package (object):
    """ Generic Package class
        A Package is a working container for one or more output
        packages for a specific package type (e.g., nuget) """

    def __init__(self, version, arts):
        super(Package, self).__init__()
        self.version = version
        self.arts = arts
        # These may be overwritten by specific sub-classes:
        self.artifacts = arts.artifacts
        # Staging path, filled in later.
        self.stpath = None
        self.kv = {'version': version}
        self.files = dict()

    def add_file(self, file):
        self.files[file] = True

    def build(self):
        """ Build package output(s), return a list of paths "
        to built packages """
        raise NotImplementedError

    def cleanup(self):
        """ Optional cleanup routine for removing temporary files, etc. """
        pass

    def render(self, fname, destpath='.'):
        """ Render template in file fname and save to destpath/fname,
        where destpath is relative to stpath """

        outf = os.path.join(self.stpath, destpath, fname)

        if not os.path.isdir(os.path.dirname(outf)):
            os.makedirs(os.path.dirname(outf), 0o0755)

        with open(os.path.join('templates', fname), 'r') as tf:
            tmpl = Template(tf.read())
        with open(outf, 'w') as of:
            of.write(tmpl.substitute(self.kv))

        self.add_file(outf)

    def copy_template(self, fname, target_fname=None, destpath='.'):
        """ Copy template file to destpath/fname
        where destpath is relative to stpath """

        if target_fname is None:
            target_fname = fname
        outf = os.path.join(self.stpath, destpath, target_fname)

        if not os.path.isdir(os.path.dirname(outf)):
            os.makedirs(os.path.dirname(outf), 0o0755)

        shutil.copy(os.path.join('templates', fname), outf)

        self.add_file(outf)

    def apply_mappings(self):
        """ Applies a list of Mapping to match and extract files from
            matching artifacts. If any of the listed Mappings can not be
            fulfilled an exception is raised. """

        assert self.mappings
        assert len(self.mappings) > 0

        for m in self.mappings:

            artifact = None
            for a in self.arts.artifacts:
                found = True

                for attr in m.attributes:
                    if attr[0] == '!':
                        # Require attribute NOT to match
                        origattr = attr
                        attr = attr[1:]

                        if attr in a.info and \
                           a.info[attr] == m.attributes[origattr]:
                            found = False
                            break
                    else:
                        # Require attribute to match
                        if attr not in a.info or \
                           a.info[attr] != m.attributes[attr]:
                            found = False
                            break

                if not fnmatch(a.fname, m.fname_glob):
                    found = False

                for exclude in m.fname_excludes:
                    if exclude in a.fname:
                        found = False
                        break

                if found:
                    artifact = a
                    break

            if artifact is None:
                raise MissingArtifactError(
                    '%s: unable to find artifact with tags %s matching "%s"' %
                    (m, str(m.attributes), m.fname_glob))

            output_path = os.path.join(self.stpath, m.output_path)

            try:
                zfile.ZFile.extract(artifact.lpath, m.input_path, output_path)
#            except KeyError:
#                continue
            except Exception as e:
                raise Exception(
                    '%s: file not found in archive %s: %s. Files in archive are:\n%s' %  # noqa: E501
                    (m, artifact.lpath, e, '\n'.join(zfile.ZFile(
                        artifact.lpath).getnames())))

            # Check that the file type matches.
            if magic_mismatch(output_path, a):
                os.unlink(output_path)
                continue

        # All mappings found and extracted.

    def verify(self, path):
        """ Verify package content based on the previously defined mappings """

        missing = list()
        with zfile.ZFile(path, 'r') as zf:
            print('Verifying %s:' % path)

            # Zipfiles may url-encode filenames, unquote them before matching.
            pkgd = [unquote(x) for x in zf.getnames()]
            missing = [x for x in self.mappings if x.output_path not in pkgd]

        if len(missing) > 0:
            print(
                'Missing files in package %s:\n%s' %
                (path, '\n'.join([str(x) for x in missing])))
            print('Actual: %s' % '\n'.join(pkgd))
            return False

        print('OK - %d expected files found' % len(self.mappings))
        return True



#!/usr/bin/env python
import json
import sys
import re

# This script updates the Apache Kafka RPCs max versions.
# It reads the input from `input_file`, which should be a table
# looking like the table in `INTRODUCTION.md`.
# Should look like this (without the first space after the comment):
# | 0       | Produce                       | 12         | 10             |
# | 1       | Fetch                         | 17         | 16             |
# | 2       | ListOffsets                   | 10         | 7              |
# | 3       | Metadata                      | 13         | 12             |
# | 8       | OffsetCommit                  | 9          | 9              |
# | 9       | OffsetFetch                   | 9          | 9              |
# | 10      | FindCoordinator               | 6          | 2              |
# | 11      | JoinGroup                     | 9          | 5              |
# | 12      | Heartbeat                     | 4          | 3              |
# | 13      | LeaveGroup                    | 5          | 1              |
# | 14      | SyncGroup                     | 5          | 3              |
# | 15      | DescribeGroups                | 6          | 4              |
# | 16      | ListGroups                    | 5          | 4              |
# | 17      | SaslHandshake                 | 1          | 1              |
# | 18      | ApiVersions                   | 4          | 3              |
# | 19      | CreateTopics                  | 7          | 4              |
# | 20      | DeleteTopics                  | 6          | 1              |
# | 21      | DeleteRecords                 | 2          | 1              |
# | 22      | InitProducerId                | 5          | 4              |
# | 23      | OffsetForLeaderEpoch          | 4          | 2              |
# | 24      | AddPartitionsToTxn            | 5          | 0              |
# | 25      | AddOffsetsToTxn               | 4          | 0              |
# | 26      | EndTxn                        | 5          | 1              |
# | 28      | TxnOffsetCommit               | 5          | 3              |
# | 29      | DescribeAcls                  | 3          | 1              |
# | 30      | CreateAcls                    | 3          | 1              |
# | 31      | DeleteAcls                    | 3          | 1              |
# | 32      | DescribeConfigs               | 4          | 1              |
# | 33      | AlterConfigs                  | 2          | 2              |
# | 36      | SaslAuthenticate              | 2          | 1              |
# | 37      | CreatePartitions              | 3          | 0              |
# | 42      | DeleteGroups                  | 2          | 1              |
# | 43      | ElectLeaders                  | 2          | 2              |
# | 44      | IncrementalAlterConfigs       | 1          | 1              |
# | 47      | OffsetDelete                  | 0          | 0              |
# | 50      | DescribeUserScramCredentials  | 0          | 0              |
# | 51      | AlterUserScramCredentials     | 0          | 0              |
# | 68      | ConsumerGroupHeartbeat        | 1          | 1              |
# | 69      | ConsumerGroupDescribe         | 1          | 0              |
# | 71      | GetTelemetrySubscriptions     | 0          | 0              |
# | 72      | PushTelemetry                 | 0          | 0              |
#
# Output will be the same with max versions updated
# Should pass Apache Kafka root folder as first argument and the input file
# as second argument.
ak_folder = sys.argv[1]
input_file = sys.argv[2]

if len(sys.argv) != 3:
    print("Usage: python3 update_rpcs_max_versions.py <kafka_folder> "
          "<input_file>")
    sys.exit(1)

with open(input_file, 'r') as input:
    lines = input.readlines()
    max_first_column = 0
    max_second_column = 0
    max_third_column = 0
    apis = []
    for line in lines:
        line = re.sub('^\\s*\\|\\s*', '', line)
        pipe_char = line.find('|')
        max_first_column = max(max_first_column, pipe_char)
        api_num = int(line[0:pipe_char])
        line = line[pipe_char + 1:]
        line = re.sub('^\\s*', '', line)
        pipe_char = line.find('|')
        max_second_column = max(max_second_column, pipe_char)
        api = line[0:pipe_char].strip()
        line = line[pipe_char + 1:].lstrip()
        pipe_char = line.find('|')
        max_third_column = max(max_third_column, pipe_char)
        rest = line[pipe_char + 1:].strip()
        apis.append((api_num, api, rest))

    for api_num, api, rest in apis:
        with open(f'{ak_folder}/clients/src/main/resources/common/message/'
                  f'{api}Request.json',
                  'r') as f:
            text = f.readlines()
            text = "".join([line for line in text
                            if '#' not in line and '//' not in line])
            json_object = json.loads(text)
            max_version = json_object["validVersions"].split("-")[-1]
            print('| ', end='')
            print(str(api_num).ljust(max_first_column), end='')
            print('| ', end='')
            print(api.ljust(max_second_column), end='')
            print('| ', end='')
            print(str(max_version).ljust(max_third_column) + '|', end='')
            print(f' {rest}')



#!/usr/bin/env python3
#
# Create self-contained static-library tar-ball package
#

import os
import tempfile
import shutil
import subprocess
from packaging import Package, Mapping


class StaticPackage (Package):
    """ Create a tar-ball with self-contained static libraries.
        These are later imported into confluent-kafka-go. """

    # Make sure gssapi (cyrus-sasl) is not linked, since that is a
    # dynamic linkage, by specifying negative match '!extra': 'gssapi'.
    # Except for on OSX where cyrus-sasl is always available, and
    # Windows where it is never linked.
    #
    # Match statically linked artifacts (which are included in 'all' builds)
    mappings = [
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'dist': 'centos8',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/include/librdkafka/rdkafka.h',
                'rdkafka.h'),
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'dist': 'centos8',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/share/doc/librdkafka/LICENSES.txt',
                'LICENSES.txt'),

        # glibc linux static lib and pkg-config file
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'dist': 'centos8',
                 'lnk': 'all',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka-static.a',
                'librdkafka_glibc_linux_amd64.a'),
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'dist': 'centos8',
                 'lnk': 'all',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/lib/pkgconfig/rdkafka-static.pc',
                'librdkafka_glibc_linux_amd64.pc'),

        # glibc linux arm64 static lib and pkg-config file
        Mapping({'arch': 'arm64',
                 'plat': 'linux',
                 'dist': 'centos8',
                 'lnk': 'all',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka-static.a',
                'librdkafka_glibc_linux_arm64.a'),
        Mapping({'arch': 'arm64',
                 'plat': 'linux',
                 'dist': 'centos8',
                 'lnk': 'all',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/lib/pkgconfig/rdkafka-static.pc',
                'librdkafka_glibc_linux_arm64.pc'),

        # glibc linux s390x static lib and pkg-config file
        Mapping({'arch': 's390x',
                 'plat': 'linux',
                 'dist': 'centos8',
                 'lnk': 'all',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka-static.a',
                'librdkafka_glibc_linux_s390x.a'),
        Mapping({'arch': 's390x',
                 'plat': 'linux',
                 'dist': 'centos8',
                 'lnk': 'all',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/lib/pkgconfig/rdkafka-static.pc',
                'librdkafka_glibc_linux_s390x.pc'),

        # musl linux static lib and pkg-config file
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'dist': 'alpine',
                 'lnk': 'all',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka-static.a',
                'librdkafka_musl_linux_amd64.a'),
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'dist': 'alpine',
                 'lnk': 'all',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/lib/pkgconfig/rdkafka-static.pc',
                'librdkafka_musl_linux_amd64.pc'),

        # musl linux arm64 static lib and pkg-config file
        Mapping({'arch': 'arm64',
                 'plat': 'linux',
                 'dist': 'alpine',
                 'lnk': 'all',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka-static.a',
                'librdkafka_musl_linux_arm64.a'),
        Mapping({'arch': 'arm64',
                 'plat': 'linux',
                 'dist': 'alpine',
                 'lnk': 'all',
                 '!extra': 'gssapi'},
                'librdkafka.tgz',
                './usr/local/lib/pkgconfig/rdkafka-static.pc',
                'librdkafka_musl_linux_arm64.pc'),

        # osx x64 static lib and pkg-config file
        Mapping({'arch': 'x64',
                 'plat': 'osx',
                 'lnk': 'all'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka-static.a',
                'librdkafka_darwin_amd64.a'),
        Mapping({'arch': 'x64',
                 'plat': 'osx',
                 'lnk': 'all'},
                'librdkafka.tgz',
                './usr/local/lib/pkgconfig/rdkafka-static.pc',
                'librdkafka_darwin_amd64.pc'),

        # osx arm64 static lib and pkg-config file
        Mapping({'arch': 'arm64',
                 'plat': 'osx',
                 'lnk': 'all'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka-static.a',
                'librdkafka_darwin_arm64.a'),
        Mapping({'arch': 'arm64',
                 'plat': 'osx',
                 'lnk': 'all'},
                'librdkafka.tgz',
                './usr/local/lib/pkgconfig/rdkafka-static.pc',
                'librdkafka_darwin_arm64.pc'),

        # win static lib and pkg-config file (mingw)
        Mapping({'arch': 'x64',
                 'plat': 'win',
                 'dist': 'mingw',
                 'lnk': 'static'},
                'librdkafka.tgz',
                './lib/librdkafka-static.a', 'librdkafka_windows.a'),
        Mapping({'arch': 'x64',
                 'plat': 'win',
                 'dist': 'mingw',
                 'lnk': 'static'},
                'librdkafka.tgz',
                './lib/pkgconfig/rdkafka-static.pc',
                'librdkafka_windows.pc'),
    ]

    def __init__(self, version, arts):
        super(StaticPackage, self).__init__(version, arts)

    def cleanup(self):
        if os.path.isdir(self.stpath):
            shutil.rmtree(self.stpath)

    def build(self, buildtype):
        """ Build single package for all artifacts. """

        self.stpath = tempfile.mkdtemp(prefix="out-", dir=".")

        self.apply_mappings()

        print('Tree extracted to %s' % self.stpath)

        # After creating a bare-bone layout, create a tarball.
        outname = "librdkafka-static-bundle-%s.tgz" % self.version
        print('Writing to %s in %s' % (outname, self.stpath))
        subprocess.check_call("(cd %s && tar cvzf ../%s .)" %
                              (self.stpath, outname),
                              shell=True)

        return outname



#!/usr/bin/env python3
#
# librdkafka test trivup app module
#
# Requires:
#  trivup python module
#  gradle in your PATH

from trivup.trivup import App, UuidAllocator
from trivup.apps.ZookeeperApp import ZookeeperApp
from trivup.apps.KafkaBrokerApp import KafkaBrokerApp
from trivup.apps.KerberosKdcApp import KerberosKdcApp
from trivup.apps.OauthbearerOIDCApp import OauthbearerOIDCApp

import os
import json


class LibrdkafkaTestApp(App):
    """ Sets up and executes the librdkafka regression tests.
        Assumes tests are in the current directory.
        Must be instantiated after ZookeeperApp and KafkaBrokerApp """

    def __init__(self, cluster, version, conf=None,
                 tests=None, scenario="default"):
        super(LibrdkafkaTestApp, self).__init__(cluster, conf=conf)

        self.appid = UuidAllocator(self.cluster).next(self, trunc=8)
        self.autostart = False
        self.local_tests = True
        self.test_mode = conf.get('test_mode', 'bare')
        self.version = version

        # Generate test config file
        conf_blob = list()
        self.security_protocol = 'PLAINTEXT'

        f, self.test_conf_file = self.open_file('test.conf', 'perm')
        f.write('broker.address.family=v4\n'.encode('ascii'))
        f.write(('test.sql.command=sqlite3 rdktests\n').encode('ascii'))
        f.write('test.timeout.multiplier=2\n'.encode('ascii'))

        sparse = conf.get('sparse_connections', None)
        if sparse is not None:
            f.write('enable.sparse.connections={}\n'.format(
                sparse).encode('ascii'))

        if version.startswith('0.9') or version.startswith('0.8'):
            conf_blob.append('api.version.request=false')
            conf_blob.append('broker.version.fallback=%s' % version)

        # SASL (only one mechanism supported at a time)
        mech = self.conf.get('sasl_mechanisms', '').split(',')[0]
        if mech != '':
            conf_blob.append('sasl.mechanisms=%s' % mech)
            if mech == 'PLAIN' or mech.find('SCRAM-') != -1:
                self.security_protocol = 'SASL_PLAINTEXT'
                # Use first user as SASL user/pass
                for up in self.conf.get('sasl_users', '').split(','):
                    u, p = up.split('=')
                    conf_blob.append('sasl.username=%s' % u)
                    conf_blob.append('sasl.password=%s' % p)
                    break

            elif mech == 'OAUTHBEARER':
                self.security_protocol = 'SASL_PLAINTEXT'
                oidc = cluster.find_app(OauthbearerOIDCApp)
                if oidc is not None:
                    conf_blob.append('sasl.oauthbearer.method=%s\n' %
                                     oidc.conf.get('sasl_oauthbearer_method'))
                    conf_blob.append('sasl.oauthbearer.client.id=%s\n' %
                                     oidc.conf.get(
                                         'sasl_oauthbearer_client_id'))
                    conf_blob.append('sasl.oauthbearer.client.secret=%s\n' %
                                     oidc.conf.get(
                                         'sasl_oauthbearer_client_secret'))
                    conf_blob.append('sasl.oauthbearer.extensions=%s\n' %
                                     oidc.conf.get(
                                         'sasl_oauthbearer_extensions'))
                    conf_blob.append('sasl.oauthbearer.scope=%s\n' %
                                     oidc.conf.get('sasl_oauthbearer_scope'))
                    conf_blob.append('sasl.oauthbearer.token.endpoint.url=%s\n'
                                     % oidc.conf.get('valid_url'))
                    self.env_add('VALID_OIDC_URL', oidc.conf.get('valid_url'))
                    self.env_add(
                        'INVALID_OIDC_URL',
                        oidc.conf.get('badformat_url'))
                    self.env_add(
                        'EXPIRED_TOKEN_OIDC_URL',
                        oidc.conf.get('expired_url'))
                    self.env_add(
                        'EXPIRED_TOKEN_OIDC_URL',
                        oidc.conf.get('expired_url'))
                    self.env['OAUTHBEARER_CLIENT_PRIVATE_KEY'] = \
                        oidc.conf['sasl_oauthbearer_client_private_key_path']
                    self.env['OAUTHBEARER_CLIENT_PRIVATE_KEY_ENCRYPTED'] = \
                        oidc.conf[
                            'sasl_oauthbearer_client_private_key_encrypted_path']  # noqa: E501
                    self.env['OAUTHBEARER_CLIENT_PRIVATE_KEY_PASSWORD'] = \
                        oidc.conf[
                            'sasl_oauthbearer_client_private_key_password']
                else:
                    conf_blob.append(
                        'enable.sasl.oauthbearer.unsecure.jwt=true\n')
                    if not self.conf.get('sasl_oauthbearer_config'):
                        conf_blob.append(
                            'sasl.oauthbearer.config=scope=requiredScope'
                            ' principal=admin\n')
                    else:
                        conf_blob.append(
                            'sasl.oauthbearer.config=%s\n' %
                            self.conf.get('sasl_oauthbearer_config'))

            elif mech == 'GSSAPI':
                self.security_protocol = 'SASL_PLAINTEXT'
                kdc = cluster.find_app(KerberosKdcApp)
                if kdc is None:
                    self.log(
                        'WARNING: sasl_mechanisms is GSSAPI set but no '
                        'KerberosKdcApp available: client SASL config will '
                        'be invalid (which might be intentional)')
                else:
                    self.env_add('KRB5_CONFIG', kdc.conf['krb5_conf'])
                    self.env_add('KRB5_KDC_PROFILE', kdc.conf['kdc_conf'])
                    principal, keytab = kdc.add_principal(
                        self.name,
                        conf.get('advertised_hostname', self.node.name))
                    conf_blob.append('sasl.kerberos.service.name=%s' %
                                     self.conf.get('sasl_servicename',
                                                   'kafka'))
                    conf_blob.append('sasl.kerberos.keytab=%s' % keytab)
                    conf_blob.append(
                        'sasl.kerberos.principal=%s' %
                        principal.split('@')[0])

            else:
                self.log(
                    'WARNING: FIXME: SASL %s client config not written to %s: unhandled mechanism' %  # noqa: E501
                    (mech, self.test_conf_file))

        # SSL config
        if getattr(cluster, 'ssl', None) is not None:
            ssl = cluster.ssl
            ssl_intermediate_ca = cluster.conf.get(
                'ssl_intermediate_ca', False)

            key = ssl.create_cert('librdkafka%s' % self.appid,
                                  through_intermediate=ssl_intermediate_ca)

            conf_blob.append('ssl.ca.location=%s' % ssl.ca['pem'])
            conf_blob.append('ssl.certificate.location=%s' % key['pub']['pem'])
            conf_blob.append('ssl.key.location=%s' % key['priv']['pem'])
            conf_blob.append('ssl.key.password=%s' % key['password'])

            # Some tests need fine-grained access to various cert files,
            # set up the env vars accordingly.
            for k, v in ssl.ca.items():
                self.env_add('SSL_ca_{}'.format(k), v)

            for k, v in ssl.unused_ca.items():
                self.env['SSL_unused_ca_{}'.format(k)] = v

            self.env['SSL_all_cas_pem'] = ssl.all_cas['pem']
            self.env['SSL_client_auth'] = cluster.conf.get(
                'ssl_client_auth', 'required')

            # Set envs for all generated keys so tests can find them.
            for k, v in key.items():
                if isinstance(v, dict):
                    for k2, v2 in v.items():
                        # E.g. "SSL_priv_der=path/to/librdkafka-priv.der"
                        self.env_add('SSL_{}_{}'.format(k, k2), v2)
                else:
                    self.env_add('SSL_{}'.format(k), v)

            if 'SASL' in self.security_protocol:
                self.security_protocol = 'SASL_SSL'
            else:
                self.security_protocol = 'SSL'

        # Define bootstrap brokers based on selected security protocol
        self.dbg('Using client security.protocol=%s' % self.security_protocol)
        all_listeners = (
            ','.join(
                cluster.get_all(
                    'advertised.listeners',
                    '',
                    KafkaBrokerApp))).split(',')
        bootstrap_servers = ','.join(
            [x for x in all_listeners if x.startswith(self.security_protocol)])
        if len(bootstrap_servers) == 0:
            bootstrap_servers = all_listeners[0]
            self.log(
                'WARNING: No eligible listeners for security.protocol=%s in %s: falling back to first listener: %s: tests will fail (which might be the intention)' %  # noqa: E501
                (self.security_protocol, all_listeners, bootstrap_servers))

        self.bootstrap_servers = bootstrap_servers

        conf_blob.append('bootstrap.servers=%s' % bootstrap_servers)
        conf_blob.append('security.protocol=%s' % self.security_protocol)

        f.write(('\n'.join(conf_blob)).encode('ascii'))
        f.close()

        self.env_add('TEST_SCENARIO', scenario)
        self.env_add('RDKAFKA_TEST_CONF', self.test_conf_file)
        test_kafka_version = version
        if 'TEST_KAFKA_VERSION' not in os.environ:
            self.env_add('TEST_KAFKA_VERSION', test_kafka_version, False)
        self.env_add('TRIVUP_ROOT', cluster.instance_path())

        if self.test_mode != 'bash':
            self.test_report_file = self.mkpath('test_report', pathtype='perm')
            self.env_add('TEST_REPORT', self.test_report_file)

            if tests is not None:
                self.env_add('TESTS', ','.join(tests))

    def finalize_env(self):
        self.env_add(
            'KAFKA_PATH',
            self.cluster.get_all(
                'destdir',
                '',
                KafkaBrokerApp)[0],
            False)

        zookeeper = self.cluster.get_all(
            'address',
            '',
            ZookeeperApp)
        if len(zookeeper):
            self.env_add(
                'ZK_ADDRESS',
                zookeeper[0],
                False)
        self.env_add('BROKERS', self.cluster.bootstrap_servers(), False)

        # Provide a HTTPS REST endpoint for the HTTP client tests.
        self.env_add(
            'RD_UT_HTTP_URL',
            'https://jsonplaceholder.typicode.com/users',
            False)

        # Per broker env vars
        for b in [x for x in self.cluster.apps if isinstance(
                x, KafkaBrokerApp)]:
            self.env_add('BROKER_ADDRESS_%d' % b.appid,
                         ','.join([x for x in
                                   b.conf['listeners'].split(',')
                                   if x.startswith(self.security_protocol)]),
                         False)
            # Add each broker pid as an env so they can be killed
            # indivdidually.
            self.env_add('BROKER_PID_%d' % b.appid, str(b.proc.pid), False)
            # JMX port, if available
            jmx_port = b.conf.get('jmx_port', None)
            if jmx_port is not None:
                self.env_add(
                    'BROKER_JMX_PORT_%d' %
                    b.appid, str(jmx_port), False)

    def start_cmd(self):
        self.finalize_env()

        extra_args = list()
        if not self.local_tests:
            extra_args.append('-L')
        if self.conf.get('args', None) is not None:
            extra_args.append(self.conf.get('args'))
        extra_args.append('-E')
        return './run-test.sh -p%d -K %s %s' % (
            int(self.conf.get('parallel', 5)), ' '.join(extra_args),
            self.test_mode)

    def report(self):
        if self.test_mode == 'bash':
            return None

        try:
            with open(self.test_report_file, 'r') as f:
                res = json.load(f)
        except Exception as e:
            self.log(
                'Failed to read report %s: %s' %
                (self.test_report_file, str(e)))
            return {'root_path': self.root_path(), 'error': str(e)}
        return res

    def deploy(self):
        pass



#!/usr/bin/env python3
#
# Create NuGet package
#

import os
import tempfile
import shutil
import subprocess
from packaging import Package, Mapping


class NugetPackage (Package):
    """ All platforms, archs, et.al, are bundled into one set of
        NuGet output packages: "main", redist and symbols """

    # See .semamphore/semaphore.yml for where these are built.
    mappings = [
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'lnk': 'std'},
                'librdkafka.tgz',
                './usr/local/include/librdkafka/rdkafka.h',
                'build/native/include/librdkafka/rdkafka.h'),
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'lnk': 'std'},
                'librdkafka.tgz',
                './usr/local/include/librdkafka/rdkafkacpp.h',
                'build/native/include/librdkafka/rdkafkacpp.h'),
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'lnk': 'std'},
                'librdkafka.tgz',
                './usr/local/include/librdkafka/rdkafka_mock.h',
                'build/native/include/librdkafka/rdkafka_mock.h'),

        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'lnk': 'std'},
                'librdkafka.tgz',
                './usr/local/share/doc/librdkafka/README.md',
                'README.md'),
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'lnk': 'std'},
                'librdkafka.tgz',
                './usr/local/share/doc/librdkafka/CONFIGURATION.md',
                'CONFIGURATION.md'),
        Mapping({'arch': 'x64',
                 'plat': 'osx',
                 'lnk': 'all'},
                'librdkafka.tgz',
                './usr/local/share/doc/librdkafka/LICENSES.txt',
                'LICENSES.txt'),

        # OSX x64
        Mapping({'arch': 'x64',
                 'plat': 'osx'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka.dylib',
                'runtimes/osx-x64/native/librdkafka.dylib'),
        # OSX arm64
        Mapping({'arch': 'arm64',
                 'plat': 'osx'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka.1.dylib',
                'runtimes/osx-arm64/native/librdkafka.dylib'),

        # Linux glibc centos8 x64 with GSSAPI
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'dist': 'centos8',
                 'lnk': 'std'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka.so.1',
                'runtimes/linux-x64/native/librdkafka.so'),
        # Linux glibc centos8 x64 without GSSAPI (no external deps)
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'dist': 'centos8',
                 'lnk': 'all'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka.so.1',
                'runtimes/linux-x64/native/centos8-librdkafka.so'),
        # Linux glibc centos8 arm64 without GSSAPI (no external deps)
        Mapping({'arch': 'arm64',
                 'plat': 'linux',
                 'dist': 'centos8',
                 'lnk': 'all'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka.so.1',
                'runtimes/linux-arm64/native/librdkafka.so'),
        # Linux glibc centos8 s390x without GSSAPI (no external deps)
        Mapping({'arch': 's390x',
                 'plat': 'linux',
                 'dist': 'centos8',
                 'lnk': 'all'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka.so.1',
                'runtimes/linux-s390x/native/librdkafka.so'),

        # Linux musl alpine x64 without GSSAPI (no external deps)
        Mapping({'arch': 'x64',
                 'plat': 'linux',
                 'dist': 'alpine',
                 'lnk': 'all'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka.so.1',
                'runtimes/linux-x64/native/alpine-librdkafka.so'),
        # Linux musl alpine arm64 without GSSAPI (no external deps)
        Mapping({'arch': 'arm64',
                 'plat': 'linux',
                 'dist': 'alpine',
                 'lnk': 'all'},
                'librdkafka.tgz',
                './usr/local/lib/librdkafka.so.1',
                'runtimes/linux-arm64/native/alpine-librdkafka.so'),

        # Common Win runtime
        Mapping({'arch': 'x64',
                 'plat': 'win'},
                'msvcr140.zip',
                'vcruntime140.dll',
                'runtimes/win-x64/native/vcruntime140.dll'),
        Mapping({'arch': 'x64',
                 'plat': 'win'},
                'msvcr140.zip',
                'msvcp140.dll', 'runtimes/win-x64/native/msvcp140.dll'),

        # matches x64 librdkafka.redist.zip
        Mapping({'arch': 'x64',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/x64/Release/librdkafka.dll',
                'runtimes/win-x64/native/librdkafka.dll'),
        Mapping({'arch': 'x64',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/x64/Release/librdkafkacpp.dll',
                'runtimes/win-x64/native/librdkafkacpp.dll'),
        Mapping({'arch': 'x64',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/x64/Release/libcrypto-3-x64.dll',
                'runtimes/win-x64/native/libcrypto-3-x64.dll'),
        Mapping({'arch': 'x64',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/x64/Release/libssl-3-x64.dll',
                'runtimes/win-x64/native/libssl-3-x64.dll'),
        Mapping({'arch': 'x64',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/x64/Release/zlib1.dll',
                'runtimes/win-x64/native/zlib1.dll'),
        Mapping({'arch': 'x64',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/x64/Release/zstd.dll',
                'runtimes/win-x64/native/zstd.dll'),
        Mapping({'arch': 'x64',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/x64/Release/libcurl.dll',
                'runtimes/win-x64/native/libcurl.dll'),
        # matches x64 librdkafka.redist.zip, lib files
        Mapping({'arch': 'x64',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/lib/v142/x64/Release/librdkafka.lib',
                'build/native/lib/win/x64/win-x64-Release/v142/librdkafka.lib'  # noqa: E501
                ),
        Mapping({'arch': 'x64',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/lib/v142/x64/Release/librdkafkacpp.lib',
                'build/native/lib/win/x64/win-x64-Release/v142/librdkafkacpp.lib'  # noqa: E501
                ),

        Mapping({'arch': 'x86',
                 'plat': 'win'},
                'msvcr140.zip',
                'vcruntime140.dll',
                'runtimes/win-x86/native/vcruntime140.dll'),
        Mapping({'arch': 'x86',
                 'plat': 'win'},
                'msvcr140.zip',
                'msvcp140.dll', 'runtimes/win-x86/native/msvcp140.dll'),

        # matches Win32 librdkafka.redist.zip
        Mapping({'arch': 'x86',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/Win32/Release/librdkafka.dll',
                'runtimes/win-x86/native/librdkafka.dll'),
        Mapping({'arch': 'x86',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/Win32/Release/librdkafkacpp.dll',
                'runtimes/win-x86/native/librdkafkacpp.dll'),
        Mapping({'arch': 'x86',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/Win32/Release/libcrypto-3.dll',
                'runtimes/win-x86/native/libcrypto-3.dll'),
        Mapping({'arch': 'x86',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/Win32/Release/libssl-3.dll',
                'runtimes/win-x86/native/libssl-3.dll'),

        Mapping({'arch': 'x86',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/Win32/Release/zlib1.dll',
                'runtimes/win-x86/native/zlib1.dll'),
        Mapping({'arch': 'x86',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/Win32/Release/zstd.dll',
                'runtimes/win-x86/native/zstd.dll'),
        Mapping({'arch': 'x86',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/bin/v142/Win32/Release/libcurl.dll',
                'runtimes/win-x86/native/libcurl.dll'),

        # matches Win32 librdkafka.redist.zip, lib files
        Mapping({'arch': 'x86',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/lib/v142/Win32/Release/librdkafka.lib',
                'build/native/lib/win/x86/win-x86-Release/v142/librdkafka.lib'  # noqa: E501
                ),
        Mapping({'arch': 'x86',
                 'plat': 'win'},
                'librdkafka.redist*',
                'build/native/lib/v142/Win32/Release/librdkafkacpp.lib',
                'build/native/lib/win/x86/win-x86-Release/v142/librdkafkacpp.lib'  # noqa: E501
                )
    ]

    def __init__(self, version, arts):
        if version.startswith('v'):
            version = version[1:]  # Strip v prefix
        super(NugetPackage, self).__init__(version, arts)

    def cleanup(self):
        if os.path.isdir(self.stpath):
            shutil.rmtree(self.stpath)

    def build(self, buildtype):
        """ Build single NuGet package for all its artifacts. """

        # NuGet removes the prefixing v from the version.
        vless_version = self.kv['version']
        if vless_version[0] == 'v':
            vless_version = vless_version[1:]

        self.stpath = tempfile.mkdtemp(prefix="out-", suffix="-%s" % buildtype,
                                       dir=".")

        self.render('librdkafka.redist.nuspec')
        self.copy_template('librdkafka.redist.targets',
                           destpath=os.path.join('build', 'native'))
        self.copy_template('librdkafka.redist.props',
                           destpath='build')

        # Generate template tokens for artifacts
        for a in self.arts.artifacts:
            if 'bldtype' not in a.info:
                a.info['bldtype'] = 'release'

            a.info['variant'] = '%s-%s-%s' % (a.info.get('plat'),
                                              a.info.get('arch'),
                                              a.info.get('bldtype'))
            if 'toolset' not in a.info:
                a.info['toolset'] = 'v142'

        # Apply mappings and extract files
        self.apply_mappings()

        print('Tree extracted to %s' % self.stpath)

        # After creating a bare-bone nupkg layout containing the artifacts
        # and some spec and props files, call the 'nuget' utility to
        # make a proper nupkg of it (with all the metadata files).
        subprocess.check_call("./nuget.sh pack %s -BasePath '%s' -NonInteractive" %  # noqa: E501
                              (os.path.join(self.stpath,
                                            'librdkafka.redist.nuspec'),
                               self.stpath), shell=True)

        return 'librdkafka.redist.%s.nupkg' % vless_version



#!/usr/bin/env python3
import os
import sys
import re
import subprocess
import signal
curr_dir = os.path.dirname(os.path.realpath(__file__))
os.chdir(curr_dir)

# This test script accepts the same arguments as run-test.sh, and
# the same environment variables as well. Additionally, it reads the following
# environment variables and can run tests in batches and for a certain
# number of iterations, especially useful when running tests with valgrind or
# to debug infrequent errors.
#
# TESTS: A comma-separated list of test numbers or test names. In case
#        they're numeric there's no need to pad them with zeros. There can be
#        ranges of tests, open ranges and they're sorted and deduplicated.
#        If not set, it defaults to all tests in the current directory.
#        Example: TESTS=1,2,3-5,2-,11
#
# TESTS_BATCH_SIZE: The number of tests to run in each batch. If not set, it
#                   defaults to the number of tests in TESTS.
#
# TEST_ITERATIONS: The number of iterations to run the tests. If not set, it
#                  defaults to 1. It continues iterating only if there are no
#                  errors. When "-a" is set it stops after the first error
#                  in any batch, when it isn't set it stops after the
#                  first error in any iteration.

# Max number of tests
max_automatic_tests = 1000
# Higher test number
max_test = 0
tests = set()
for f in os.listdir(curr_dir):
    m = re.match(r'^(\d{4}).*\.c$', f)
    if m:
        test_num = int(m.group(1))
        if test_num < max_automatic_tests:
            tests.add(test_num)
max_test = max(tests)
# Tests batch size, defaults to len(TESTS).
batch_size = os.environ.get('TESTS_BATCH_SIZE', None)
if batch_size and batch_size.isnumeric():
    batch_size = int(batch_size)

# Number of iterations for each batch, it continues iterating only
# if there are no errors, even when "-a" isn't set
test_iterations = os.environ.get('TEST_ITERATIONS', "1")
test_iterations = test_iterations.isnumeric() and int(test_iterations) or 1


def valid_test_number(test_number):
    return test_number >= 0 and test_number <= max_test


def test_number_to_str(test_number):
    return str(test_number).rjust(4, '0')


def test_numbers_to_str(test_numbers):
    return list(map(test_number_to_str, test_numbers))


def extract_tests(tests):
    test_ranges = [test_range for test_range in tests.strip().split(',')
                   if len(test_range.strip()) > 0]
    all_tests = []
    for test_range in test_ranges:

        if '-' in test_range:
            # 13-15 or 13- or -13
            start, end = test_range.split('-')
            start, end = start.strip(), end.strip()

            if start.isnumeric() and end.isnumeric():
                start, end = int(start), int(end)
                if valid_test_number(start) and valid_test_number(
                        end) and start <= end:
                    all_tests.extend(
                        test_numbers_to_str(range(start, end + 1)))

            elif start.isnumeric() and end == '':
                start = int(start)
                if valid_test_number(start):
                    all_tests.extend(
                        test_numbers_to_str(range(start, max_test + 1)))
            elif start == '' and end.isnumeric():
                end = int(end)
                if valid_test_number(end):
                    all_tests.extend(
                        test_numbers_to_str(range(0, end + 1)))

        elif test_range.isnumeric():
            # 13 or 0013
            all_tests.append(test_number_to_str(test_range))
        else:
            # A specific test like 0045_subscribe_update_racks_mock
            all_tests.append(test_range)

    return list(sorted(set(all_tests)))


if 'TESTS' in os.environ:
    TESTS = extract_tests(os.environ['TESTS'])
else:
    TESTS = test_numbers_to_str(range(0, max_test + 1))

if batch_size is None:
    batch_size = len(TESTS)

args = sys.argv[1:]

exit_on_first_error = '-a' in args


def run_tests():
    first_error = None
    for i in range(test_iterations):
        for j in range(0, len(TESTS), batch_size):
            interrupted = False
            TESTS_BATCH = ','.join(TESTS[j:j + batch_size])
            print(f"Running tests: {TESTS_BATCH}, iteration: {i + 1}",
                  file=sys.stderr)
            p = subprocess.Popen(['./run-test.sh', '-D'] + args,
                                 env={**os.environ, 'TESTS': TESTS_BATCH},
                                 start_new_session=True)
            try:
                p.communicate()
                if p.returncode != 0 and first_error is None:
                    first_error = p.returncode
                    if exit_on_first_error:
                        return first_error
            except BaseException:
                first_error = 1
                interrupted = True
                return first_error
            finally:
                if interrupted:
                    print('Terminating process group...', file=sys.stderr)
                    os.killpg(p.pid, signal.SIGTERM)
                    try:
                        p.wait(10)
                    except subprocess.TimeoutExpired:
                        os.killpg(p.pid, signal.SIGKILL)
                        p.wait(10)
        if first_error is not None:
            # Stop running more iterations if there was an error
            return first_error
    return first_error


first_error = run_tests()
print('End of run-test-batches', file=sys.stderr)
if first_error:
    sys.exit(first_error)
